/*
 * hping.c
 * Copyright (C) 1998 Salvatore Sanfilippo - antirez - <md5330@mclink.it>
 * 						      <antirez@seclab.com>
 * This source is covered by the GNU GPL
 * compile under linux. (Do you want make a porting?, mail me please)
 *
 * Intesis SECURITY LAB            Phone: +39-02-671563.1
 * Via Settembrini, 35             Fax: +39-02-66981953
 * I-20124 Milano  ITALY           Email: antirez@seclab.com
 *
 */

#define RELEASE_VERSION "0.70 - linux"
#define RELEASE_DATE "Tue Dec 22 13:14:43 MET 1998"

#define _BSD_SOURCE

#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>
#include <unistd.h>

#ifdef DONTUSELIBPCAP
  #include <sys/ioctl.h>
  #ifdef __GLIBC__	/* please, if this #ifdef doesn't work */
			/* with some Linux distr. mail me. */
    #include <netinet/if_ether.h>	/* tested on Debian 2.0 */
  #else
    #include <linux/if_ether.h>		/* tested on RedHat 4.1 */
  #endif /* __GLIBC__ */
  #include <net/if.h>
#else
  #include <pcap.h>
#endif /* DONTUSELIBPCAP */

#ifndef DONTUSELIBPCAP
 #ifdef __linux__
  #ifdef __GLIBC__
   #include <netinet/if_ether.h>
  #else
   #include <linux/if_ether.h>
  #endif /* __GLIBC__ */
 #endif /* __linux__ */
#include <net/if.h>
#include <sys/ioctl.h>
#endif /* DONTUSELIBPCAP */

#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>

/* requests status table stuffs */
#define TABLESIZE	40	/* TABLESIZE 0 == floating point exception */
#define S_SENT		0
#define S_RECV		1

/* usefull defines */
#ifndef TRUE
#define TRUE	1
#define FALSE	0
#endif

#ifdef  DONTUSELIBPCAP
#ifndef IFNAMSIZ
#define IFNAMSIZ 16
#endif
#endif  /* DONTUSELIBPCAP */

#ifndef TH_X
#define	TH_X 0x40	/* unused X tcp flag */
#endif
#ifndef TH_Y
#define TH_Y 0x80	/* unused Y tcp flag */
#endif

/* header size of some physical layer type */
#define PPPHDR_SIZE	0
#define ETHHDR_SIZE	sizeof(struct ethhdr)
#define LOHDR_SIZE	sizeof(struct ethhdr)
#define UNKNOWNHDR_SIZE	0

/* packet size (physical header size + ip header + tcp header + 0 data bytes) */
#define LINK_PACKETSIZE	( linkhdr_size +	\
			  sizeof(struct iphdr)+ \
			  sizeof(struct tcphdr) )
/* ip packet size */
#define IP_PACKETSIZE	( sizeof(struct iphdr) + sizeof(struct tcphdr) )

/* absolute offsets */
#define	ABS_OFFSETLINK	0
#define ABS_OFFSETIP	linkhdr_size
#define ABS_OFFSETTCP	( linkhdr_size + sizeof(struct iphdr) )
#define ABS_OFFSETICMP	( linkhdr_size + sizeof(struct iphdr) )

/* ip realtive offsets */
#define IP_OFFSETIP	0
#define IP_OFFSETTCP	sizeof(struct iphdr)

/* defaults and misc */
#define DEFAULT_SENDINGWAIT 1	/* wait 1 sec. between sending each packets */
#define DEFAULT_DPORT	    0
#define DEFAULT_INITSPORT  -1	/* alias random */
#define DEFAULT_COUNT      -1	/* alias forever */
#define DEFAULT_TTL	   64
#define DEFAULT_SRCWINSIZE 512

#define BIND_NONE	0
#define BIND_DPORT	1
#define BIND_TTL	2
#define DEFAULT_BIND	BIND_DPORT
	
/* fragmentation macros */
#define MF ((unsigned short)0x2000)	/* more fragments */
#define NF ((unsigned short)0x0000)	/* no more fragments */
	
/* globals
*/
#ifndef DONTUSELIBPCAP
pcap_t	*pcapfp;
char	errbuf[PCAP_ERRBUF_SIZE];
struct pcap_pkthdr hdr;
#endif
unsigned int tcp_th_flags = 0;
int	sockpacket, sockraw,			/* sockets fd */
	finsent = 0, rstrecv = 0,		/* fin sent & reset received */	
	sending_wait = DEFAULT_SENDINGWAIT,	/* see DEFAULT_SENDINGWAIT def*/
	linkhdr_size,				/* physical layer header size */
	opt_numeric	= FALSE,
	opt_icmpgethostname = TRUE,
	opt_quiet	= FALSE,
	opt_relid	= FALSE,
	fragmentation	= FALSE,
	set_mf		= FALSE,
	src_ttl		= DEFAULT_TTL,
	dport		= DEFAULT_DPORT,
	sport,
	initsport = DEFAULT_INITSPORT,
	src_winsize = DEFAULT_SRCWINSIZE,
	tcp_thoff = -1,				/* sizeof(struct tcphdr) >> 2 */
	count		= DEFAULT_COUNT,
	ctrlzbind	= DEFAULT_BIND;
char	targetname[1024],			/* target hostname */
	targetstraddr[1024],
	ifname[1024] = {'\0'},			/* interface name */
	ifstraddr[1024],			/* interface address */
	spoofaddr[1024];			/* spoof address */
struct sockaddr_in local, remote;

struct delaytable_element {
	int seq;
	int sec;
	long usec;
	int status;
};

volatile struct delaytable_element delaytable[TABLESIZE];
int	delaytable_index = 0;

/* protos
*/
int	parse_options(int, char**);		/* parse command line options */
int	get_if_name(void);			/* get interface name & addr. */
int	get_linkhdrsize(char*);			/* get link layer hdr size */
int	open_sockpacket(void);			/* open SOCK_PACKET socket */
int	close_sockpacket(int);			/* close SOCK_PACKET socket */
int	open_sockraw(void);			/* open SOCK_RAW socket */
void	sendfin(int);				/* send one FIN */
void	ip_sender (char*, char*, char*, unsigned int, int, unsigned short);
void	waitresets(void);			/* wait RST replies */
void	print_statistics(int);			/* print statistics */
void	usage(void);				/* shows a short help */
void	showversion(void);			/* shows version informations */
void	resolver(struct sockaddr*, char*);	/* resolver */
void	icmpunreach_log(char*, unsigned short); /* logs icmp ureached */
void	icmptimeexc_log(char*, unsigned short); /* logs icmp time exc */
long	get_utime(void);			/* return current usec */
u_short cksum(u_short *buf, int nwords);	/* compute 16bit checksum */
void	incdestport(int sid);			/* inc destination port */
char	*get_hostname(char*);			/* get hostname */
#ifndef DONTUSELIBPCAP
int	pcap_recv(char* , unsigned int);	/* libpcap based recv */
#endif

/* main
*/
int main(int argc, char **argv)
{
	char setflags[1024] = {'\0'};
	int c;

	if ( parse_options(argc, argv) == -1 ) usage();

	/* get interface's name and address */
	if ( get_if_name() == -1 ) {
		printf("[main] no such device\n");
		exit(1);
	}

	/* open sockpacket socket and raw socket */
	sockraw = open_sockraw();
	if (sockraw == -1) {
		printf("[main] can't open raw socket\n");
		exit(1);
	}
	sockpacket = open_sockpacket();
	if (sockpacket == -1) {
		printf("[main] can't open packet socket\n");
		exit(1);
	}

	/* get physical layer header size */
	if ( get_linkhdrsize(ifname) == -1 ) {
		printf("[main] physical layer header size unknown\n");
		exit(1);
	}

	if (spoofaddr[0] == '\0')
		resolver((struct sockaddr*)&local, ifstraddr);
	else
		resolver((struct sockaddr*)&local, spoofaddr);
	resolver((struct sockaddr*)&remote, targetname);

	srand(time(NULL));

	/* set initial source port */
	if (initsport == -1)
		initsport = sport = 1024 + (rand() % 2000);
	else
		sport = initsport;

	for (c = 0; c < TABLESIZE; c++)
		delaytable[c].seq = -1;
	
	strncpy(targetstraddr, inet_ntoa(remote.sin_addr),
		sizeof(targetstraddr));

	if (tcp_th_flags & TH_RST) strcat(setflags, "R");
	if (tcp_th_flags & TH_SYN) strcat(setflags, "S");
	if (tcp_th_flags & TH_ACK) strcat(setflags, "A");
	if (tcp_th_flags & TH_FIN) strcat(setflags, "F");
	if (tcp_th_flags & TH_PUSH) strcat(setflags, "P");
	if (tcp_th_flags & TH_URG) strcat(setflags, "U");
	if (tcp_th_flags & TH_X) strcat(setflags, "X");
	if (tcp_th_flags & TH_Y) strcat(setflags, "Y");
	if (setflags[0] == '\0') strcat(setflags, "no flags are");

	printf("HPING %s (%s %s): %s set, %d data bytes\n", targetname, ifname,
		targetstraddr, setflags, IP_PACKETSIZE);

	signal(SIGALRM, sendfin);
	if (ctrlzbind != BIND_NONE)
		signal(SIGTSTP, incdestport);
	signal(SIGINT, print_statistics);
	signal(SIGTERM, print_statistics);
	kill(getpid(), SIGALRM);

	waitresets();	/* the main loop */

	return 0;
}

int parse_options(int argc, char **argv)
{
	char c;
	int y = 1;

	if (argc < 2) return -1;

	while(argv[y]) {
		if (argv[y][0] != '-') {
			strncpy(targetname, argv[y], 1024);
			break;
		}
	y++;
	}

	while ( (c = getopt(argc, argv,
			"c:i:nqI:hvp:s:t:w:a:O:FSRPAUXYfxrzZ")) != EOF )
	{
		switch(c)
		{
			case '?':
				return -1;
			case 'c':
				count = atoi(optarg);
				break;
			case 'i':
				sending_wait = atoi(optarg);
				break;
			case 'n':
				opt_numeric = TRUE;
				break;
			case 'q':
				opt_quiet = TRUE;
				break;
			case 'I':
				strncpy (ifname, optarg, 1024);
				break;
			case 'h':
				return -1;
			case 'v':
				showversion();
			case 'p':
				dport = atoi(optarg);
				break;
			case 's':
				initsport = atoi(optarg);
				break;
			case 't':
				src_ttl = atoi(optarg);
				break;
			case 'w':
				src_winsize = atoi(optarg);
				break;
			case 'a':
				strncpy (spoofaddr, optarg, 1024);
				break;
			case 'F':
				tcp_th_flags |= TH_FIN;
				break;
			case 'S':
				tcp_th_flags |= TH_SYN;
				break;
			case 'R':
				tcp_th_flags |= TH_RST;
				break;
			case 'P':
				tcp_th_flags |= TH_PUSH;
				break;
			case 'A':
				tcp_th_flags |= TH_ACK;
				break;
			case 'U':
				tcp_th_flags |= TH_URG;
				break;
			case 'X':
				tcp_th_flags |= TH_X;
				break;
			case 'Y':
				tcp_th_flags |= TH_Y;
				break;
			case 'f':
				fragmentation = TRUE;
				break;
			case 'x':
				set_mf = TRUE;
				break;
			case 'O':
				tcp_thoff = atoi(optarg);
				break;
			case 'r':
				opt_relid = TRUE;
				break;
			case 'z':
				ctrlzbind = BIND_TTL;
				break;
			case 'Z':
				ctrlzbind = BIND_NONE;
		}

	if (argv[y] == NULL) return -1;		/* missing target host */
	if (opt_numeric == TRUE) opt_icmpgethostname = FALSE;
	}

	/* force some conditions */

	if (count <= 0 && count != -1) {
		printf("[parse_options] count must > 0\n");
		exit(1);
	}

	if (sending_wait <= 0) {
		printf("[prase_options] bad timing interval\n");
		exit(1);
	}
	return 1;
}

int get_if_name(void)
{
	int fd;
	struct ifconf ifc;
	struct ifreq ibuf[16], ifr, *ifrp, *ifend;
	struct sockaddr_in sa;

	if ( (fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		perror("[get_if_name] socket()");
		return -1;
	}

	bzero((void*)ibuf, sizeof(ibuf));	/* reset buffer */
	ifc.ifc_len = sizeof ibuf;		/* set buffer size... */
	ifc.ifc_buf = (caddr_t) ibuf;		/* and buffer pointer */

	/* gets interfaces list */
	if ( ioctl(fd, SIOCGIFCONF, (char*)&ifc) == -1 ||
	     ifc.ifc_len < sizeof(struct ifreq)		) {
		perror("[get_if_name] ioctl()");
		return -1;
	}

	/* ifrp points to buffer and ifend points to buffer's end */
	ifrp = ibuf;
	ifend = (struct ifreq*) ((char*)ibuf + ifc.ifc_len);

	for (; ifrp < ifend; ifrp++) {
		strncpy(ifr.ifr_name, ifrp->ifr_name, sizeof(ifr.ifr_name));

		if ( ioctl(fd, SIOCGIFFLAGS, (char*)&ifr) == -1) {
			perror("[get_if_name] ioctl()");
			return -1;
		}

		if ( !(ifr.ifr_flags & IFF_UP) ) continue;    /* if down */
		if ( strstr(ifr.ifr_name, "lo") ) continue;   /* if loopback */
		if ( ifname[0] != '\0' )		      /* opt -I set */
			if ( !strstr(ifr.ifr_name, ifname) )  /* don't match */
				continue;

		/* interface found */
		strncpy(ifname, ifr.ifr_name, 1024);
		memcpy(&sa, &(ifrp->ifr_addr), sizeof(struct sockaddr_in));
		strncpy(ifstraddr, inet_ntoa(sa.sin_addr), 1024);
		return 0;
	}
	/* interface not found, use 'lo' */
	strncpy(ifname, "lo", 1024);
	strncpy(ifstraddr, "127.0.0.1", 1024);
	return 0;
}

#ifdef DONTUSELIBPCAP
int get_linkhdrsize(char *ifname)
{
	if ( strstr(ifname, "ppp") )
	{
		linkhdr_size = PPPHDR_SIZE;
		return 0;
	}
	else if ( strstr(ifname, "eth") )
	{
		linkhdr_size = ETHHDR_SIZE;
		return 0;
	}
	else if ( strstr(ifname, "lo") )
	{
		linkhdr_size = LOHDR_SIZE;
		return 0;
	}
	else
		return 1;
}
#else /* DONTUSELIBPCAP */
int get_linkhdrsize(char *ifname)
{
	switch(pcap_datalink(pcapfp))
	{
		case DLT_EN10MB:
			linkhdr_size = ETHHDR_SIZE;
			break;

		case DLT_NULL:
			linkhdr_size = 0;
			break;

		case DLT_PPP:
#ifdef __linux__
			linkhdr_size = 0;
#else
			linkhdr_size = 4;
#endif /* __linux__ */
		break;

		case 0x0c: /* linux PPP */
			linkhdr_size = 0;
			break;

		default:
			printf("%d\n", pcap_datalink(pcapfp));
			printf("[get_linkhdrsize] Unknown datalink layer\n");
			return -1;
	}
	return 0;
}
#endif /* DONTUSELIBPCAP */

int open_sockraw()
{
	int s;

	s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (s == -1) {
		perror("[open_sockraw] socket()");
		return -1;
	}

	return s;
}

#ifdef DONTUSELIBPCAP
int open_sockpacket()
{
	int s;
	struct ifreq ifr;

	s = socket(AF_INET, SOCK_PACKET, htons(ETH_P_IP));
	if (s == -1) {
		perror("[open_sockpacket] socket()");
		return -1;
	}

	strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
	if ( ioctl(s, SIOCGIFFLAGS, &ifr) == -1) {    /* get interface flags */
		perror("[open_sockpacket] ioctl()");
		return -1;
	}
//	ifr.ifr_flags |= IFF_PROMISC;	/* set promiscuose mode */
	if ( ioctl(s, SIOCSIFFLAGS, &ifr) == -1) {    /* set interface flags */
		perror("[open_sockpacket] ioctl()");
		return -1;
	}

	return s;
}
#else /* DONTUSELIBPCAP */
int open_sockpacket()
{
	pcapfp = pcap_open_live(ifname, 99999, 0, 9000, errbuf);
	if (pcapfp == NULL) {
		printf("[open_sockpacket] pcap_open_live: %s\n", errbuf);
		return -1;
	}

	return 0;
}
#endif /* DONTUSELIBPCAP */

#ifdef DONTUSELIBPCAP
int close_sockpacket(int s)
{
	struct ifreq ifr;

	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
	if ( ioctl(s, SIOCGIFFLAGS, &ifr) == -1) {		/* get flags */
		perror("[open_sockpacket] ioctl()");
		return -1;
	}

	if (ifr.ifr_flags & IFF_PROMISC) {
		ifr.ifr_flags ^= IFF_PROMISC;	   /* unset promiscuose mode */
		if ( ioctl(s, SIOCSIFFLAGS, &ifr) == -1) {	/* set flags */
			perror("[open_sockpacket] ioctl()");
			return -1;
		}
	}

	return close(s);
}
#else /* DONTUSELIBPCAP */
int close_sockpacket(int s)
{
	return 0;
}
#endif /* DONTUSELIBPCAP */

void resolver (struct sockaddr * addr, char *hostname)
{
	struct  sockaddr_in *address;
	struct  hostent     *host;

	address = (struct sockaddr_in *)addr;

	bzero((char *)address, sizeof(struct sockaddr_in));
	address->sin_family = AF_INET;
	address->sin_addr.s_addr = inet_addr(hostname);

	if ( (int)address->sin_addr.s_addr == -1) {
		host = gethostbyname(hostname);
		if (host) {
			bcopy( host->h_addr,
			(char *)&address->sin_addr,host->h_length);
		} else {
			perror("[resolver] Could not resolve address");
			exit(1);
		}
	}
}

void ip_sender (char* src, char* dst, char *data, unsigned int datalen,
		int more_fragments, unsigned short fragoff)
{
	char			*packet;
	int			result, packetsize;
	struct iphdr		*ip;

	packetsize = sizeof(struct iphdr) + datalen;
	if ( (packet = malloc(packetsize)) == NULL) {
		perror("[ip_sender] malloc()");
		return;
	}

	bzero(packet, packetsize);
	ip = (struct iphdr*) packet;

	/* copy src and dst address */
	bcopy(src, &ip->saddr, sizeof(ip->saddr));
	bcopy(dst, &ip->daddr, sizeof(ip->daddr));

	/* ip header */
	ip->version	= 4;
	ip->ihl		= sizeof(struct iphdr) >> 2;
	ip->tos		= 0;
	ip->tot_len	= htons(packetsize);
	ip->id		= htons(getpid() & 255);
	ip->frag_off	|= htons(more_fragments);
	ip->frag_off	|= htons(fragoff >> 3); /* shift three flags bit */
	ip->ttl		= src_ttl;
	ip->protocol	= 6; /* tcp */
	ip->check	= 0; /* always computed by the kernel */

	/* copies data */
	bcopy(data, packet+sizeof(struct iphdr), datalen);
	
	result = sendto(sockraw, packet, packetsize, 0,
		(struct sockaddr*)&remote, sizeof(remote));
	
	if (result == -1 && errno != EINTR) {
		perror("[ip_sender] sendto");
		if (close(sockraw) == -1)
			perror("[ipsender] close(sockraw)");
		if (close_sockpacket(sockpacket) == -1)
			perror("[ipsender] close(sockpacket)");
		exit(1);
	}

	free(packet);
}

void sendfin (int signal_id)
{
	int			tcphdr_size;
	char			*packet_tcphdr;
	struct tcphdr		*tcp;
	struct tcp_pseudohdr
	{
		struct in_addr saddr;
		struct in_addr daddr;
		u_char zero;
		u_char protocol;
		u_short lenght;
		struct tcphdr tcpheader;
	} pseudoheader;

	tcphdr_size = sizeof(struct tcphdr);
	if ( (packet_tcphdr = malloc(tcphdr_size)) == NULL) {
		perror("[sendfin] malloc()");
		return;
	}
	tcp =  (struct tcphdr*) packet_tcphdr;
	
	signal(SIGALRM, sendfin);
	bzero(packet_tcphdr, tcphdr_size);

	/* tcp header */
	tcp->th_dport	= htons(dport);
	tcp->th_sport	= htons(sport);
	tcp->th_seq	= htonl(rand());
	tcp->th_ack	= htonl(0);
	if (tcp_thoff == -1)
		tcp->th_off	= sizeof(struct tcphdr) >> 2;
	else
		tcp->th_off	= tcp_thoff;
	tcp->th_win	= htons(src_winsize);
	tcp->th_flags	= tcp_th_flags;

	/* compute checksum */
	bzero(&pseudoheader, 12+sizeof(struct tcphdr));
	pseudoheader.saddr.s_addr = local.sin_addr.s_addr;
	pseudoheader.daddr.s_addr = remote.sin_addr.s_addr;
	pseudoheader.protocol = 6; /* tcp */
	pseudoheader.lenght = htons(sizeof(struct tcphdr));
	bcopy((char*) tcp, (char*) &pseudoheader.tcpheader,
		sizeof(struct tcphdr));
	tcp->th_sum = cksum((u_short*) &pseudoheader, 12+sizeof(struct tcphdr));

	if (!fragmentation)
	{
		if (!set_mf)
			ip_sender((char*)&local.sin_addr,
					(char*)&remote.sin_addr,
					packet_tcphdr,
					tcphdr_size, NF, 0);
		else
			ip_sender((char*)&local.sin_addr,
					(char*)&remote.sin_addr,
					packet_tcphdr,
					tcphdr_size, MF, 0);
	}
	else
	{
		ip_sender((char*)&local.sin_addr,
				(char*)&remote.sin_addr,
				packet_tcphdr,
				16, MF, 0);
		ip_sender((char*)&local.sin_addr,
				(char*)&remote.sin_addr,
				packet_tcphdr+16,
				4, NF, 16);
	}

	/* adds this fin in delaytable */
	delaytable[delaytable_index % TABLESIZE].seq = sport - initsport;
	delaytable[delaytable_index % TABLESIZE].sec = time(NULL);
	delaytable[delaytable_index % TABLESIZE].usec = get_utime();
	delaytable[delaytable_index % TABLESIZE].status = S_SENT;
	delaytable_index++;	/* inc. index */
	
	finsent++;		/* inc. sent fin counter */
	
	if (count != -1 && count == finsent)	/* count reached */
		print_statistics(0);

	alarm(sending_wait);
	sport++;		/* inc. source port */
}


/* this isn't a good  general checksum algorithm, but i don't need to pad tcp's
   segments, so it's good for me. */
u_short cksum(u_short *buf, int nwords)
{
	unsigned long sum;
	u_short *w = buf;
	
	for (sum = 0; nwords > 0; nwords-=2)
		sum += *w++;
	
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);
	return ~sum;
}

void	waitresets(void)
{
	char	packet[LINK_PACKETSIZE+16];
	char recvflags[1024];
	struct	iphdr	*ip	= (struct iphdr*) (packet+ABS_OFFSETIP);
	struct	tcphdr	*tcp	= (struct tcphdr*) (packet+ABS_OFFSETTCP);
	struct	icmphdr	*icmp	= (struct icmphdr*) (packet+ABS_OFFSETICMP);
	struct	iphdr	*icmp_iph = (struct iphdr*) (packet+ABS_OFFSETICMP+8);

	bzero((void*)packet, sizeof(packet));	/* ?????? */

	while(1)
	{
		char	src_addr[1024],
			dst_addr[1024];
		int	src_port, dst_port, size, rstseq, winsize,
			sec_delay, usec_delay, status;
		float	ms_delay;
		struct	in_addr src, dst;

#ifdef DONTUSELIBPCAP		
		size = recv(sockpacket, packet, LINK_PACKETSIZE+16, 0);
#else
		size = pcap_recv(packet, LINK_PACKETSIZE+16);
#endif /* DONTUSELIBPCAP */

		if (size == -1) {
			if (errno != EINTR)
				print_statistics(-1);
			else
				continue;
		}

		bcopy( &(ip->saddr), &src, sizeof(struct in_addr) );
		bcopy( &(ip->daddr), &dst, sizeof(struct in_addr) );
		strncpy(src_addr, inet_ntoa(src), 1024);
		strncpy(dst_addr, inet_ntoa(dst), 1024);
		src_port = ntohs(tcp->th_sport);
		dst_port = ntohs(tcp->th_dport);	

		if ( ip->protocol == IPPROTO_TCP &&
		     !memcmp(&ip->saddr, &remote.sin_addr, sizeof(ip->saddr)) &&
		     !memcmp(&ip->daddr, &local.sin_addr, sizeof(ip->daddr)) &&
		     src_port == dport
		)
		{
			int i,
			fin_tablepos = -1,
			ip_id,
			backup_id,
			rel_id = FALSE;
			static int	last_seq = 0,
					last_id = -1;

			rstseq = dst_port - initsport;
			winsize = ntohs(tcp->th_win);
			backup_id = ip_id = htons(ip->id);

			recvflags[0] = '\0';
			if (tcp->th_flags & TH_RST) strcat(recvflags, "R");
			if (tcp->th_flags & TH_SYN) strcat(recvflags, "S");
			if (tcp->th_flags & TH_ACK) strcat(recvflags, "A");
			if (tcp->th_flags & TH_FIN) strcat(recvflags, "F");
			if (tcp->th_flags & TH_PUSH) strcat(recvflags, "P");
			if (tcp->th_flags & TH_URG) strcat(recvflags, "U");
			if (tcp->th_flags & TH_X) strcat(recvflags, "X");
			if (tcp->th_flags & TH_Y) strcat(recvflags, "Y");
			if (recvflags[0] == '\0') strcat(recvflags, "none");
			
			for (i=0; i<TABLESIZE; i++)
				if (delaytable[i].seq == rstseq) {
					fin_tablepos = i;
					break;
				}

			if (fin_tablepos != -1)
			{
				sec_delay = time(NULL) - delaytable[fin_tablepos].sec;
				usec_delay = get_utime() - delaytable[fin_tablepos].usec;
				ms_delay = (sec_delay * 1000) + ((float)usec_delay / 1000);
				status = delaytable[fin_tablepos].status;
				delaytable[fin_tablepos].status = S_RECV;
			}
			else
			{
				ms_delay = 0;	/* not in table.. */
				status = 0;	/* we don't know if it's DUP */
			}

			/* relativize packets id */
			if (opt_relid)
			{
				int seq_diff;

				if (last_id == -1) {
					last_id = ip_id;
					last_seq = rstseq;
				}
				else
					if ( (seq_diff=(rstseq-last_seq)) > 0)
					{
						if (last_id > ip_id) /* rew */
							ip_id = ((65535-last_id)
							    + ip_id)/seq_diff;
						else
							ip_id = (ip_id-last_id)
								/seq_diff;
					last_id = backup_id;
					last_seq = rstseq;
					rel_id = TRUE;
					}
			}

			printf("%d bytes from %s: ", size, src_addr);
			printf("flags=%s ", recvflags);
			printf("seq=%d ", rstseq);
			printf("ttl=%d ", ip->ttl);
			printf("id");
			if (rel_id == TRUE)
				printf("=+");
			else
				putchar('=');
			printf("%d ", ip_id);
			printf("win=%d ", winsize);
			printf("time=%.1f ms", ms_delay);
			
//			printf("%d bytes from %s: flags=%s seq=%d ttl=%d id=%d win=%d time=%.1f ms", size, src_addr, recvflags, rstseq, ip->ttl, ip_id, winsize, ms_delay);
			if (status == S_RECV)
				printf(" ( DUP! ) ");
			putchar('\n');

			fflush(stdout);
			rstrecv++;
		}

		if ( ip->protocol == IPPROTO_ICMP )
		{
			if ( icmp->type == 3 && /* Dest. Unreachable */
				!memcmp(&icmp_iph->daddr,
					&remote.sin_addr,
					sizeof(ip->daddr)) &&
				!memcmp(&ip->daddr,
					&local.sin_addr,
					sizeof(ip->daddr))
			)
				icmpunreach_log(src_addr, icmp->code);

			else if (icmp->type == 11 && /* Time exceeded */
				 !memcmp(&icmp_iph->daddr,
					 &remote.sin_addr,
					 sizeof(ip->daddr)) &&
				 !memcmp(&ip->daddr,
					 &local.sin_addr,
					 sizeof(ip->daddr))
			)
				icmptimeexc_log(src_addr, icmp->code);
		}
	}
}

void	icmptimeexc_log(char *src_addr, unsigned short icmp_code)
{
	switch(icmp_code)
	{
		case ICMP_EXC_TTL:
			printf("TTL 0 during transit from %s", src_addr);
			break;
		case ICMP_EXC_FRAGTIME:
			printf("TTL 0 during reassembly from %s", src_addr);
	}
	if (opt_icmpgethostname) {
		fflush(stdout);
		printf(" (%s)", get_hostname(src_addr));
	}
	putchar('\n');
}
		
void	icmpunreach_log(char *src_addr, unsigned short icmp_code)
{
	switch(icmp_code)
	{
		case ICMP_HOST_UNREACH:
			printf("Host Unreachable from %s", src_addr);
			break;
		case ICMP_PORT_UNREACH:
			printf("Port Unreachable from %s", src_addr);
			break;
		default:
			printf("ICMP Unreachable type %d from %s", icmp_code, src_addr);
			break;
	}
	if (opt_icmpgethostname) {
		fflush(stdout);
		printf(" (%s)", get_hostname(src_addr));
	}
	putchar('\n');
}
	
void	print_statistics(int signal_id)
{
	unsigned int lossrate;
	
	close_sockpacket(sockpacket);

	if (rstrecv > 0)
		lossrate = 100 - ((rstrecv*100)/finsent);
	else
		if (!finsent)
			lossrate = 0; /* no packets sent, no packets loss... */
		else
			lossrate = 100; /* packets sent... nadaz returned */

	printf("\n--- %s hping statistic ---\n", targetname);
	printf("%d packets tramitted, %d packets received, %d%% packet loss\n", finsent, rstrecv, lossrate);
	exit(0);
};

void	usage(void)
{
	int	index = 0;
	char	*helpdata[] = {
"usage: hping host [options]\n",
"\tc - packet count\n",
"\ti - wait\n"
"\tn - numeric output\n",
"\tq - quiet\n",
"\tI - interface name\n",
"\tp - destination port (default 0) [ctrl+z inc, double ctrl+z dec]\n",
"\ts - base source port (default random)\n",
"\tt - ttl (default 64)\n",
"\tw - winsize (default 512)\n",
"\ta - spoof host addr|hostname\n",
"\th - show this help\n",
"\tv - show version\n",
"\t F - set FIN flag                S - set SYN flag\n",
"\t R - set RST flag                P - set PUSH flag\n",
"\t A - set ACK flag                U - set URG flag\n",
"\t X - set X unused flag (0x40)    Y = set Y unused flag (0x80)\n",
"\tf - split packets in two fragments (may pass weak acl)\n",
"\tx - set more fragments flag (maybe useless)\n",
"\tO - set fake tcp data offset (instead of tcphdrlen / 4)\n",
"\tr - relativize id field (to estimate host traffic)\n",
"\tz - bind ctrl+z to ttl\n",
"\tZ - unbind ctrl+z\n",
		NULL };

	while (helpdata[index])
	{
		printf(helpdata[index]);
		++index;
	}
	exit(0);
};

long get_utime(void)
{
	struct timeval tmptv;

	gettimeofday(&tmptv, NULL);
	return tmptv.tv_usec;
}

void showversion(void)
{
	printf("hping version %s (%s)\n", RELEASE_VERSION, RELEASE_DATE);
	printf("Salvatore Sanfilippo <antirez@seclab.com>\n");
	printf("                      <md5330@mclink.it>\n");
	exit(0);
}

void incdestport(int sid)
{
	static long sec = 0;
	static long usec = 0;
	int *p;

	if (ctrlzbind == BIND_DPORT)
		p = &dport;
	else if (ctrlzbind == BIND_TTL)
		p = &src_ttl;
	else {
		printf("problem binding ctrl+z...\n");
		return;
	}
	
	if ( (time(NULL) == sec) && ((get_utime() - usec) < 200000) )
	{
		if (*p > 0)
			(*p)-=2;
		if (*p < 0)
			*p=0;
	} else
		(*p)++;
	
	printf("\b\b\b\b\b\b\b\b\b");
	printf("%d: ", *p);
	fflush(stdout);

	sec = time(NULL);
	usec = get_utime();
	
	signal(SIGTSTP, incdestport);
}

char *get_hostname(char* addr)
{
	static char answer[1024];
	static char lastreq[1024] = {'\0'};	/* last request and */
	struct hostent *he;
	struct in_addr naddr;
	static char *last_answerp = NULL;

	printf(" get hostname..."); fflush(stdout);
	printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");

	if ( !strcmp(addr, lastreq)) return last_answerp;

	strncpy(lastreq, addr, 1024);

	inet_aton(addr, &naddr);
	he = gethostbyaddr((char*)&naddr, 4, AF_INET);

	if (he == NULL)
	{
		last_answerp = NULL;
		return NULL;
	}

	strncpy(answer, he->h_name, 1024);
	last_answerp = answer;

	return answer;
}

#ifndef DONTUSELIBPCAP
int pcap_recv(char *packet, unsigned int size)
{
	char *p = NULL;
	int pcapsize;

	while(p == NULL)
		p = (unsigned char*) pcap_next(pcapfp, &hdr);

	pcapsize = hdr.len;

	if (pcapsize < size)
		size = pcapsize;

	bcopy(p, packet, size);

	return size;
}
#endif
